@@ -19,7 +19,7 @@ from binarycpython.utils.functions import (
     Capturing,
 )
 from binarycpython.utils.ensemble import (
-    binarycDecoder,
+    BinarycDecoder,
     handle_ensemble_string_to_json,
     extract_ensemble_json_from_string,
 )
@@ -119,8 +119,6 @@ class test_run_system(unittest.TestCase):
 ### memaddr test
 #######################################################################################################################################################
 
-# TODO: Make some assertion tests in c
-
 
 class test_return_store_memaddr(unittest.TestCase):
     """
@@ -156,7 +154,7 @@ class test_return_store_memaddr(unittest.TestCase):
 #######################################################################################################################################################
 
 
-class TestEnsemble(unittest.TestCase):
+class test_ensemble_functions(unittest.TestCase):
     """
     Unittests for handling the ensemble outputs and adding those
     """
@@ -165,7 +163,7 @@ class TestEnsemble(unittest.TestCase):
         """
         init function
         """
-        super(TestEnsemble, self).__init__(*args, **kwargs)
+        super(test_ensemble_functions, self).__init__(*args, **kwargs)
 
     def test_return_persistent_data_memaddr(self):
         with Capturing() as output:
@@ -180,9 +178,6 @@ class TestEnsemble(unittest.TestCase):
         output = _binary_c_bindings.return_persistent_data_memaddr()
 
         self.assertTrue(is_capsule(output), msg="Object must be a capsule")
-        # self.assertNotEqual(
-        #     output, 0, "memory adress seems not to have a correct value"
-        # )
 
     def test_minimal_ensemble_output(self):
         with Capturing() as output:
@@ -442,14 +437,13 @@ class TestEnsemble(unittest.TestCase):
 
     #############
 
-    # def test_full_ensemble_output(self):
-    #     with Capturing() as output:
-    #         self._test_full_ensemble_output()
+    def test_full_ensemble_output(self):
+        with Capturing() as output:
+            self._test_full_ensemble_output()
 
     def _test_full_ensemble_output(self):
         """
         Function to just output the whole ensemble
-        TODO: put this one back
         """
         print(self.id())
 
@@ -466,13 +460,12 @@ class TestEnsemble(unittest.TestCase):
         #
         output_json_1 = extract_ensemble_json_from_string(output_1)
 
-        keys = json_1.keys()
+        keys = output_json_1.keys()
 
         # assert statements:
         self.assertIn("number_counts", keys)
         self.assertIn("HRD", keys)
         self.assertIn("HRD(t)", keys)
-        self.assertIn("Xyield", keys)
         self.assertIn("distributions", keys)
         self.assertIn("scalars", keys)
 
