@@ -18,14 +18,6 @@ from binarycpython.utils.useful_funcs import (
     calc_period_from_sep,
 )
 
-# class test_(unittest.TestCase):
-#     """
-#     Unittests for function
-#     """
-
-#     def test_1(self):
-#         pass
-
 
 class test_calc_period_from_sep(unittest.TestCase):
     """
@@ -105,7 +97,7 @@ class test_ragb(unittest.TestCase):
         """
 
         m = 20
-        output = ragb(m, 0.02)
+        output = ragb(m)
 
         self.assertEqual(output, 820)
 
