@@ -1,152 +1,198 @@
 """
-Module containing functions that binary_c-python uses to modify
-dictionaries.
+Module containing functions that binary_c-python uses to modify dictionaries.
 """
-import copy
+
 import collections
+from typing import Union
+
 import astropy.units as u
 import numpy as np
-from collections import (
-    OrderedDict,
-)
 
-# we need to convert keys to floats:
-# this is ~ a factor 10 faster than David's
-# recursive_change_key_to_float routine,
-# probably because this version only does
-# the float conversion, nothing else.
-def keys_to_floats(json_data):
-    # assumes nested dicts ...
-    # new_data = {}
+# Define all numerical types
 
-    # but this copies the variable type, but has some
-    # pointless copying
-    # new_data = copy.copy(json_data)
-    # new_data.clear()
+ALLOWED_NUMERICAL_TYPES = (int, float, complex, np.number)
+UNION_ALLOWED_NUMERICAL_TYPES = Union[int, float, complex, np.number]
+
+
+def keys_to_floats(input_dict: dict) -> dict:
+    """
+    Function to convert all the keys of the dictionary to float to float
+
+    we need to convert keys to floats:
+        this is ~ a factor 10 faster than David's ``recursive_change_key_to_float`` routine, probably because this version only does the float conversion, nothing else.
+
+    Args:
+        input_dict: dict of which we want to turn all the keys to float types if possible
+
+    Returns:
+        new_dict: dict of which the keys have been turned to float types where possible
+    """
 
     # this adopts the type correctly *and* is fast
-    new_data = type(json_data)()
+    new_dict = type(input_dict)()
 
-    for k, v in json_data.items():
+    for k, v in input_dict.items():
+        # convert key to a float, if we can
+        # otherwise leave as is
+        try:
+            newkey = float(k)
+        except ValueError:
+            newkey = k
+
+        # act on value(s)
         if isinstance(v, list):
-            v = [
+            # list data
+            new_dict[newkey] = [
                 keys_to_floats(item)
                 if isinstance(item, collections.abc.Mapping)
                 else item
                 for item in v
             ]
         elif isinstance(v, collections.abc.Mapping):
-            # dict, ordereddict, etc.
-            v = keys_to_floats(v)
-        try:
-            f = float(k)
-            new_data[f] = json_data[k]
-        except:
-            new_data[k] = v
-    return new_data
+            # dict, ordereddict, etc. data
+            new_dict[newkey] = keys_to_floats(v)
+        else:
+            # assume all other data are scalars
+            new_dict[newkey] = v
+
+    return new_dict
 
 
-def recursive_change_key_to_float(input_dict):
+def recursive_change_key_to_float(input_dict: dict) -> dict:
     """
     Function to recursively change the key to float
 
     This only works if the dict contains just sub-dicts or numbers/strings.
+
     Does not work with lists as values
+
+    Args:
+        input_dict: dict of which we want to turn all the keys to float types if possible
+
+    Returns:
+        new_dict: dict of which the keys have been turned to float types where possible
+
+    If input_dict is None or empty, returns an empty dict
     """
 
-    new_dict = collections.OrderedDict()  # TODO: check if this still works
+    new_dict = collections.OrderedDict()
 
-    for key in input_dict:
-        if isinstance(input_dict[key], (dict, collections.OrderedDict)):
-            try:
-                num_key = float(key)
-                new_dict[num_key] = recursive_change_key_to_float(
-                    copy.deepcopy(input_dict[key])
-                )
-            except ValueError:
-                new_dict[key] = recursive_change_key_to_float(
-                    copy.deepcopy(input_dict[key])
-                )
-        else:
-            try:
-                num_key = float(key)
-                new_dict[num_key] = input_dict[key]
-            except ValueError:
-                new_dict[key] = input_dict[key]
+    # if the input dict is None or empty, return an empty dict
+    if input_dict is None or not input_dict:
+        pass
+
+    else:
+        # dict has keys, loop over them
+        for key in input_dict:
+            if isinstance(input_dict[key], (dict, collections.OrderedDict)):
+                try:
+                    num_key = float(key)
+                    new_dict[num_key] = recursive_change_key_to_float(input_dict[key])
+                except ValueError:
+                    new_dict[key] = recursive_change_key_to_float(input_dict[key])
+            else:
+                try:
+                    num_key = float(key)
+                    new_dict[num_key] = input_dict[key]
+                except ValueError:
+                    new_dict[key] = input_dict[key]
 
     return new_dict
 
 
-def recursive_change_key_to_string(input_dict):
+def recursive_change_key_to_string(input_dict: dict, custom_format: str = "{:g}"):
     """
-    Function to recursively change the key back to a string but this time in a format that we decide
+    Function to recursively change the key back to a string but this time in a format that we decide. We'll try to turn a string key into a float key before formatting the key
+
+    Args:
+        input_dict: dict of which we want to turn all the keys to string types (with a custom format)
+        custom_format: custom format used when turning the key to strings
+
+    Returns:
+        new_dict: dict of which the keys have been turned to string types where possible
     """
 
-    new_dict = collections.OrderedDict()  # TODO: check if this still works
+    new_dict = collections.OrderedDict()
 
     for key in input_dict:
+        # Try to turn into a float
+        try:
+            string_key = float(key)
+        except ValueError:
+            string_key = key
+
+        # Turn into string with new format
+        if not isinstance(string_key, str):
+            string_key = custom_format.format(string_key)
+
+        # If dictionary type, call function again
         if isinstance(input_dict[key], (dict, collections.OrderedDict)):
-            if isinstance(key, (int, float)):
-                string_key = "{:g}".format(key)
-                new_dict[string_key] = recursive_change_key_to_string(
-                    copy.deepcopy(input_dict[key])
-                )
-            else:
-                new_dict[key] = recursive_change_key_to_string(
-                    copy.deepcopy(input_dict[key])
-                )
+            new_dict[string_key] = recursive_change_key_to_string(
+                input_dict[key], custom_format
+            )
         else:
-            if isinstance(key, (int, float)):
-                string_key = "{:g}".format(key)
-                new_dict[string_key] = input_dict[key]
-            else:
-                new_dict[key] = input_dict[key]
+            new_dict[string_key] = input_dict[key]
 
     return new_dict
 
 
-############################################################
-# code to walk a dictionary recursively based on
-# https://stackoverflow.com/questions/13687924/setting-a-value-in-a-nested-python-dictionary-given-a-list-of-indices-and-value
 def _nested_set(dic, keys, value):
+    """
+    Code to set a value of a nested dict based on a list of keys. We take into account the fact that the vallue in the dict might not be set at all by the setdefault call and the reverse looping of the keys
+
+    https://stackoverflow.com/questions/13687924/setting-a-value-in-a-nested-python-dictionary-given-a-list-of-indices-and-value
+
+    TODO: describe better
+    """
+
     for key in keys[:-1]:
         dic = dic.setdefault(key, {})
     dic[keys[-1]] = value
 
 
 def _nested_get(dic, keys):
+    """
+    Code to get a value of a nested dict based on a list of keys. We take into account the fact that the vallue in the dict might not be set at all by the setdefault call and the reverse looping of the keys
+
+    TODO: unused. Remove?
+    """
+
     for key in keys[:-1]:
         dic = dic.setdefault(key, {})
     return dic[keys[-1]]
 
 
-# function to walk through the dictionary, multiplying
-# only float values by a const
-def _recursive_normalize_floats(path, d, const, parent=None, ignore=None):
+def _recursive_normalize_floats(path, input_dict, factor, parent=None, ignore=None):
+    """
+    Function to walk through the dictionary, multiplying only float values by a factor
+    """
+
     if not parent:
-        parent = d
-    for k, v in d.items():
+        parent = input_dict
+
+    for k, v in input_dict.items():
         if ignore and k in ignore:
             continue
 
         if isinstance(v, float):
             path.append(k)
-            # must be a float, multiply by the constant
-            _nested_set(parent, path, v * const)
+            # must be a float, multiply by the factor
+            _nested_set(parent, path, v * factor)
             path.pop()
-        elif isinstance(v, str) or isinstance(v, int):
+        elif isinstance(v, (str, int)):
             path.append(k)
             # do nothing to strings or ints
             path.pop()
         elif v is None:
             path.append(k)
             path.pop()
+
         # dicts
         # note: isinstance isn't enough, we need to check the Mapping
         elif isinstance(v, collections.abc.Mapping):
             path.append(k)
             # nested dict
-            _recursive_normalize_floats(path, v, const, parent=parent)
+            _recursive_normalize_floats(path, v, factor, parent=parent)
             path.pop()
         else:
             print(
@@ -156,38 +202,39 @@ def _recursive_normalize_floats(path, d, const, parent=None, ignore=None):
             )
 
 
-def multiply_float_values(d, const, ignore=None):
+def multiply_float_values(input_dict, factor, ignore=None):
     """
-    multiply_float_values : A function to recursively multiply values of a (nested) dictionary that are floats by a constant. Nested dictionaries call this function recursively.
+    A function to recursively multiply values of a (nested) dictionary that are floats by a constant. Nested dictionaries call this function recursively.
 
     Args:
-        d = the dictionary
-        const = the constant that multiplies float values
+        input_dict: the dictionary
+        factor: the constant that multiplies float values
     """
+
     path = []
-    _recursive_normalize_floats(path, d, const, parent=d, ignore=ignore)
+    _recursive_normalize_floats(
+        path, input_dict, factor, parent=input_dict, ignore=ignore
+    )
 
 
 def subtract_dicts(dict_1: dict, dict_2: dict) -> dict:
     """
-    Function to subtract two dictionaries.
+    Function to subtract two dictionaries, i.e. ``dict_1 - dict_2``
 
     Only allows values to be either a dict or a numerical type
 
     For the overlapping keys (key name present in both dicts):
-        When the keys are of the same type:
-            - If the types are of numerical type: subtract the value at dict 2 from dict 1.
-            - If the types are both dictionaries: call this function with the subdicts
+        When the keys are of the same type: If the types are of numerical type we subtract the value at dict 2 from dict 1. If the types are both dictionaries: call this function with the subdicts
 
-        When the keys are not of the same type:
-            - if the keys are all of numerical types
+        When the keys are not of the same type: If the keys are all of numerical types we do the subtraction. If they are not numerical we raise an error.
 
     For the unique keys:
-        - if the key is from dict 1: adds the value to the new dict (be it numerical value or dict)
-        - If the key is from dict 2: Adds the negative of its value in case of numerical type.
-            if the type is a dict, the result of subtract_dicts({}, dict_2[key]) will be set
+        If the key is from dict 1: adds the value to the new dict (be it numerical value or dict)
+
+        If the key is from dict 2: Adds the negative of its value in case of numerical type. If the type is a dict, the result of ``subtract_dicts({}, dict_2[key])`` will be set
 
     If the result is 0, the key will be removed from the resulting dict.
+
     If that results in an empty dict, the dict will be removed too.
 
     Args:
@@ -195,15 +242,12 @@ def subtract_dicts(dict_1: dict, dict_2: dict) -> dict:
         dict_2: second dictionary
 
     Returns:
-        Subtracted dictionary
+        Subtracted dictionary, i.e. ``dict_1 - dict_2``
     """
 
     # Set up new dict
     new_dict = {}
 
-    # Define allowed numerical types
-    ALLOWED_NUMERICAL_TYPES = (float, int, np.float64)
-
     #
     keys_1 = dict_1.keys()
     keys_2 = dict_2.keys()
@@ -223,10 +267,8 @@ def subtract_dicts(dict_1: dict, dict_2: dict) -> dict:
             if new_dict[key] == 0:
                 del new_dict[key]
 
-        # Else, to be safe we should deepcopy them
         elif isinstance(dict_1[key], dict):
-            copy_dict = copy.deepcopy(dict_1[key])
-            new_dict[key] = copy_dict
+            new_dict[key] = dict_1[key]
         else:
             msg = "Error: using unsupported type for key {}: {}".format(
                 key, type(dict_1[key])
@@ -267,13 +309,12 @@ def subtract_dicts(dict_1: dict, dict_2: dict) -> dict:
                     del new_dict[key]
 
             else:
-                msg = "Error key: {} value: {} type: {} and key: {} value: {} type: {} are not of the same type and cannot be merged".format(
-                    key,
-                    dict_1[key],
-                    type(dict_1[key]),
-                    key,
-                    dict_2[key],
-                    type(dict_2[key]),
+                msg = "Error key: {key} value: {value1} type: {type} and key: {key} value: {value2} type: {type2} are not of the same type and cannot be merged".format(
+                    key=key,
+                    value1=dict_1[key],
+                    type=type(dict_1[key]),
+                    value2=dict_2[key],
+                    type2=type(dict_2[key]),
                 )
 
                 print(msg)
@@ -289,12 +330,10 @@ def subtract_dicts(dict_1: dict, dict_2: dict) -> dict:
                 if new_dict[key] == 0:
                     del new_dict[key]
 
-            # Else, to be safe we should deepcopy them
             elif isinstance(dict_1[key], dict):
                 new_dict[key] = subtract_dicts(dict_1[key], dict_2[key])
 
                 # Remove entry if it results in an empty dict
-                # TODO: write test to prevent empty dicts from showing up
                 if not new_dict[key]:
                     del new_dict[key]
             else:
@@ -323,6 +362,10 @@ class AutoVivificationDict(dict):
     """
 
     def __getitem__(self, item):
+        """
+        Getitem function for the autovivication dict
+        """
+
         try:
             return dict.__getitem__(self, item)
         except KeyError:
@@ -330,6 +373,10 @@ class AutoVivificationDict(dict):
             return value
 
     def __iadd__(self, other):
+        """
+        iadd function (handling the +=) for the autovivication dict.
+        """
+
         # if a value does not exist, assume it is 0.0
         try:
             self += other
@@ -358,7 +405,7 @@ def inspect_dict(
             type(input_dict[key]) (except if the value is a dict)
     """
 
-    structure_dict = collections.OrderedDict()  # TODO: check if this still works
+    structure_dict = collections.OrderedDict()
 
     #
     for key, value in input_dict.items():
@@ -375,9 +422,15 @@ def inspect_dict(
     return structure_dict
 
 
-def count_keys_recursive(input_dict):
+def count_keys_recursive(input_dict: dict) -> int:
     """
-    Function to count the total number of keys in a dictionary
+    Function to recursively count the total number of keys in a dictionary.
+
+    Args:
+        input_dict: dictionary that we want to know the total amount of keys from.
+
+    Returns:
+        local_count: total amount of keys within the input_dict.
     """
 
     local_count = 0
@@ -385,6 +438,7 @@ def count_keys_recursive(input_dict):
         local_count += 1
         if isinstance(input_dict[key], (dict, collections.OrderedDict)):
             local_count += count_keys_recursive(input_dict[key])
+
     return local_count
 
 
@@ -401,6 +455,10 @@ def merge_dicts(dict_1: dict, dict_2: dict) -> dict:
         - dictionaries will be merged by calling recursively calling this function again
         - numbers will be added
         - (opt) lists will be appended
+        - booleans are merged with logical OR
+        - identical strings are just set to the string
+        - non-identical strings are concatenated
+        - NoneTypes are set to None
         - In the case that the instances do not match: for now I will raise an error
 
     Args:
@@ -413,7 +471,7 @@ def merge_dicts(dict_1: dict, dict_2: dict) -> dict:
     """
 
     # Set up new dict
-    new_dict = collections.OrderedDict()  # TODO: check if this still necessary
+    new_dict = collections.OrderedDict()
 
     #
     keys_1 = dict_1.keys()
@@ -428,22 +486,10 @@ def merge_dicts(dict_1: dict, dict_2: dict) -> dict:
 
     # Add the unique keys to the new dict
     for key in unique_to_dict_1:
-        # If these items are ints or floats, then just put them in
-        if isinstance(dict_1[key], (float, int)):
-            new_dict[key] = dict_1[key]
-        # Else, to be safe we should deepcopy them
-        else:
-            copy_dict = copy.deepcopy(dict_1[key])
-            new_dict[key] = copy_dict
+        new_dict[key] = dict_1[key]
 
     for key in unique_to_dict_2:
-        # If these items are ints or floats, then just put them in
-        if isinstance(dict_2[key], (float, int)):
-            new_dict[key] = dict_2[key]
-        # Else, to be safe we should deepcopy them
-        else:
-            copy_dict = copy.deepcopy(dict_2[key])
-            new_dict[key] = copy_dict
+        new_dict[key] = dict_2[key]
 
     # Go over the common keys:
     for key in overlapping_keys:
@@ -451,8 +497,8 @@ def merge_dicts(dict_1: dict, dict_2: dict) -> dict:
         # If they keys are not the same, it depends on their type whether we still deal with them at all, or just raise an error
         if not type(dict_1[key]) is type(dict_2[key]):
             # Exceptions: numbers can be added
-            if isinstance(dict_1[key], (int, float, np.float64)) and isinstance(
-                dict_2[key], (int, float, np.float64)
+            if isinstance(dict_1[key], ALLOWED_NUMERICAL_TYPES) and isinstance(
+                dict_2[key], ALLOWED_NUMERICAL_TYPES
             ):
                 new_dict[key] = dict_1[key] + dict_2[key]
 
@@ -464,20 +510,48 @@ def merge_dicts(dict_1: dict, dict_2: dict) -> dict:
             ):
                 new_dict[key] = merge_dicts(dict_1[key], dict_2[key])
 
+            # string-int clash : convert both to ints and save
+            elif (
+                isinstance(dict_1[key], str)
+                and isinstance(dict_2[key], int)
+                or isinstance(dict_1[key], int)
+                and isinstance(dict_2[key], str)
+            ):
+                try:
+                    new_dict[key] = int(dict_1[key]) + int(dict_2[key])
+                except ValueError as e:
+                    msg = "{}: Failed to convert string (either '{}' or '{}') to an int".format(
+                        key, dict_1[key], dict_2[key]
+                    )
+                    raise ValueError(msg) from e
+
+            # string-float clash : convert both to floats and save
+            elif (
+                isinstance(dict_1[key], str)
+                and isinstance(dict_2[key], float)
+                or isinstance(dict_1[key], float)
+                and isinstance(dict_2[key], str)
+            ):
+                try:
+                    new_dict[key] = float(dict_1[key]) + float(dict_2[key])
+                except ValueError as e:
+                    msg = "{}: Failed to convert string (either '{}' or '{}') to an float".format(
+                        key, dict_1[key], dict_2[key]
+                    )
+                    raise ValueError(msg) from e
+
             # If the above cases have not dealt with it, then we should raise an error
             else:
-                print(
-                    "Error key: {} value: {} type: {} and key: {} value: {} type: {} are not of the same type and cannot be merged".format(
-                        key,
-                        dict_1[key],
-                        type(dict_1[key]),
-                        key,
-                        dict_2[key],
-                        type(dict_2[key]),
-                    )
+                msg = "merge_dicts error: key: {key} value: {value1} type: {type1} and key: {key} value: {value2} type: {type2} are not of the same type and cannot be merged".format(
+                    key=key,
+                    value1=dict_1[key],
+                    type1=type(dict_1[key]),
+                    value2=dict_2[key],
+                    type2=type(dict_2[key]),
                 )
-                raise ValueError
+                raise ValueError(msg)
 
+        # Here the keys are the same type
         # Here we check for the cases that we want to explicitly catch. Ints will be added,
         # floats will be added, lists will be appended (though that might change) and dicts will be
         # dealt with by calling this function again.
@@ -508,17 +582,28 @@ def merge_dicts(dict_1: dict, dict_2: dict) -> dict:
             elif isinstance(dict_1[key], dict) and isinstance(dict_2[key], dict):
                 new_dict[key] = merge_dicts(dict_1[key], dict_2[key])
 
+            # strings
+            elif isinstance(dict_1[key], str) and isinstance(dict_2[key], str):
+                if dict_1[key] == dict_2[key]:
+                    # same strings
+                    new_dict[key] = dict_1[key]
+                else:
+                    # different strings: just concatenate them
+                    new_dict[key] = dict_1[key] + dict_2[key]
+
+            # None types
+            elif dict_1[key] is None and dict_2[key] is None:
+                new_dict[key] = None
+
             else:
-                print(
-                    "Object types {}: {} ({}), {} ({}) not supported.".format(
-                        key,
-                        dict_1[key],
-                        type(dict_1[key]),
-                        dict_2[key],
-                        type(dict_2[key]),
-                    )
+                msg = "Object types {}: {} ({}), {} ({}) not supported.".format(
+                    key,
+                    dict_1[key],
+                    type(dict_1[key]),
+                    dict_2[key],
+                    type(dict_2[key]),
                 )
-                raise ValueError
+                raise ValueError(msg)
 
     #
     return new_dict
@@ -529,12 +614,9 @@ def update_dicts(dict_1: dict, dict_2: dict) -> dict:
     Function to update dict_1 with values of dict_2 in a recursive way.
 
     Behaviour:
+        When dict keys are only present in one of either: we just add the content to the new dict
 
-    When dict keys are only present in one of either:
-        - we just add the content to the new dict
-
-    When dict keys are present in both, we decide based on the value types how to combine them:
-        - value of dict2 will be taken
+        When dict keys are present in both, we decide based on the value types how to combine them: value of dict2 will be taken
 
     Args:
         dict_1: first dictionary
@@ -546,14 +628,9 @@ def update_dicts(dict_1: dict, dict_2: dict) -> dict:
     """
 
     # Set up new dict of the same type as dict_1
-    #
-    # Note: setting directly to OrderedDict fails in some cases
-    # so instead we take a (shallow) copy of dict_1 which will
-    # have the same type as dict_1, then clear it. (There must
-    # be a better way to do this...)
-    new_dict = dict_1.copy()  # OrderedDict()  # TODO: check if this still works
-    new_dict.clear()
+    new_dict = dict_1.__class__()
 
+    # Get keys
     keys_1 = dict_1.keys()
     keys_2 = dict_2.keys()
 
@@ -566,22 +643,10 @@ def update_dicts(dict_1: dict, dict_2: dict) -> dict:
 
     # Add the unique keys to the new dict
     for key in unique_to_dict_1:
-        # If these items are ints or floats, then just put them in
-        if isinstance(dict_1[key], (float, int)):
-            new_dict[key] = dict_1[key]
-        # Else, to be safe we should deepcopy them
-        else:
-            copy_dict = copy.deepcopy(dict_1[key])
-            new_dict[key] = copy_dict
+        new_dict[key] = dict_1[key]
 
     for key in unique_to_dict_2:
-        # If these items are ints or floats, then just put them in
-        if isinstance(dict_2[key], (float, int)):
-            new_dict[key] = dict_2[key]
-        # Else, to be safe we should deepcopy them
-        else:
-            copy_dict = copy.deepcopy(dict_2[key])
-            new_dict[key] = copy_dict
+        new_dict[key] = dict_2[key]
 
     # Go over the common keys:
     for key in overlapping_keys:
@@ -589,20 +654,19 @@ def update_dicts(dict_1: dict, dict_2: dict) -> dict:
         # See whether the types are actually the same
         if not type(dict_1[key]) is type(dict_2[key]):
             # Exceptions:
-            if (type(dict_1[key]) in [int, float]) and (
-                type(dict_2[key]) in [int, float]
+            if isinstance(dict_1[key], ALLOWED_NUMERICAL_TYPES) and isinstance(
+                dict_2[key], ALLOWED_NUMERICAL_TYPES
             ):
                 new_dict[key] = dict_2[key]
 
             else:
                 print(
-                    "Error key: {} value: {} type: {} and key: {} value: {} type: {} are not of the same type and cannot be merged".format(
-                        key,
-                        dict_1[key],
-                        type(dict_1[key]),
-                        key,
-                        dict_2[key],
-                        type(dict_2[key]),
+                    "Error key: {key} value: {value1} type: {type1} and key: {key} value: {value2} type: {type2} are not of the same type and cannot be merged".format(
+                        key=key,
+                        value1=dict_1[key],
+                        type1=type(dict_1[key]),
+                        value2=dict_2[key],
+                        type2=type(dict_2[key]),
                     )
                 )
                 raise ValueError
@@ -621,11 +685,20 @@ def update_dicts(dict_1: dict, dict_2: dict) -> dict:
     return new_dict
 
 
-def multiply_values_dict(input_dict, factor):
+def multiply_values_dict(input_dict: dict, factor: UNION_ALLOWED_NUMERICAL_TYPES):
     """
     Function that goes over dictionary recursively and multiplies the value if possible by a factor
 
-    If the key equals "general_info", the multiplication gets skipped
+    If the key equals "general_info", the multiplication gets skipped.
+
+    This function changes the values in-place, so the original dict is modified
+
+    Args:
+        input_dict: dictionary of which we want to multiply the values by <factor>
+        factor: factor that we want to multiply the values with
+
+    Returns:
+        multiplied_dict: dict containing the multiplied keys. This is the same object as we passed as input.
     """
 
     for key in input_dict:
@@ -639,7 +712,7 @@ def multiply_values_dict(input_dict, factor):
     return input_dict
 
 
-def custom_sort_dict(input_dict):
+def custom_sort_dict(input_dict: dict) -> dict:
     """
     Returns a dictionary that is ordered, but can handle numbers better than normal OrderedDict
 
@@ -649,6 +722,9 @@ def custom_sort_dict(input_dict):
     This is done until all the keys are sorted.
 
     All objects other than dictionary types are directly return as they are
+
+    Args:
+        input_dict: object which will be sorted (and returned as a new object) if its a dictionary, otherwise it will be returned without change.
     """
 
     # If the new input is a dictionary, then try to sort it
@@ -666,6 +742,7 @@ def custom_sort_dict(input_dict):
         # If there are multiple types, then we loop over them and do a piece wise sort
         if len(all_types_keys) > 1:
             msg = "Different types in the same dictionary key set"
+            print(msg)
 
             # Create a string repr of the type name to sort them afterwards
             str_types = {repr(el): el for el in all_types_keys}
@@ -684,9 +761,10 @@ def custom_sort_dict(input_dict):
             sorted_keys = sorted(keys)
 
         for key in sorted_keys:
-            new_dict[key] = custom_sort_dict(copy.deepcopy(input_dict[key]))
+            new_dict[key] = custom_sort_dict(input_dict[key])
 
         return new_dict
+
     return input_dict
 
 
@@ -728,3 +806,69 @@ def filter_dict_through_values(arg_dict: dict, filter_list: list) -> dict:
             new_dict[key] = arg_dict[key]
 
     return new_dict
+
+
+def prepare_dict(global_dict: dict, list_of_sub_keys: list) -> None:
+    """
+    Function that makes sure that the global dict is prepared to have a value set there.
+    This dictionary will store values and factors for the distribution functions,
+    so that they don't have to be calculated each time.
+
+    Args:
+        global_dict: globally accessible dictionary where factors are stored in
+        list_of_sub_keys: List of keys that must become be(come) present in the global_dict
+    """
+
+    internal_dict_value = global_dict
+
+    # This loop almost mimics a recursive loop into the dictionary.
+    # It checks whether the first key of the list is present, if not; set it with an empty dict.
+    # Then it overrides itself to be that (new) item, and goes on to do that again, until the list
+    # exhausted
+    for k in list_of_sub_keys:
+        # If the sub key doesnt exist then make an empty dict
+        if not internal_dict_value.get(k, None):
+            internal_dict_value[k] = {}
+        internal_dict_value = internal_dict_value[k]
+
+
+def set_opts(opts: dict, newopts: dict) -> dict:
+    """
+    Function to take a default dict and override it with newer values.
+
+    # TODO: consider changing this to just a dict.update
+
+    Args:
+        opts: dictionary with default values
+        newopts: dictionary with new values
+
+    Returns:
+        returns an updated dictionary
+    """
+
+    if newopts:
+        for opt in newopts.keys():
+            if opt in opts.keys():
+                opts[opt] = newopts[opt]
+
+    return opts
+
+
+def normalize_dict(result_dict: dict) -> dict:
+    """
+    Function to normalise a dictionary by summing all the values and dividing each term by the total. Designed for dictionary containing only positive values.
+
+    Args:
+        result_dict: dictionary where values should be positive number objects
+
+    Returns:
+        normalized_dict: dictionary where the values are normalised to sum to 1
+    """
+
+    normalized_dict = {}
+
+    sum_result = sum(list(result_dict.values()))
+    for key in result_dict.keys():
+        normalized_dict[key] = result_dict[key] / sum_result
+
+    return normalized_dict
