@@ -28,8 +28,8 @@ def run_system(**kwargs):
         - run_system(M_1=10, parse_function=fancy_parsing_function)
 
     Tasks:
-        - TODO: Expand functionality.
         - TODO: Notify user when an unknown keyword is passed.
+        - TODO: Centralise the input checking routines here, abstract them and have the population object and this one use the same
 
     All the arguments known to binary_c can be passed to this function as kwargs.
     Several extra arguments can be passed through the kwargs:
@@ -65,19 +65,18 @@ def run_system(**kwargs):
     binary_c_args = {}
 
     # Check which binary_c arguments have been passed and put them into a dict
-    for key in kwargs:
+    for key, value in kwargs.items():
         if key in available_binary_c_arg_keywords:
-            binary_c_args[key] = kwargs[key]
+            binary_c_args[key] = value
 
         elif any(
-            [
-                True
-                if (key.startswith(param[:-2]) and len(param[:-2]) < len(key))
-                else False
-                for param in special_keywords
-            ]
+            bool((key.startswith(param[:-2]) and len(param[:-2]) < len(key)))
+            # True
+            # if (key.startswith(param[:-2]) and len(param[:-2]) < len(key))
+            # else False
+            for param in special_keywords
         ):
-            binary_c_args[key] = kwargs[key]
+            binary_c_args[key] = value
 
         # Notify user when this key wont be used
         else:
